SystemOrganization addCategory: #TypeFeedback!

Array variableSubclass: #TFMonomorphicCache
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TypeFeedback'!

!TFMonomorphicCache commentStamp: 'mha 6/7/2007 15:08' prior: 0!
An instance of this class represents an "inline cache" object, i.e., an object that caches, for a given send site, the most recently resolved concrete method implementation. The class is a subclass of Array. Its instances are 5 slots long:

	1:	selector			-	the selector sent at this send site (Symbol)
	2:	targetClass		-	the most recently resolved concrete target class (Class)
	3:	targetMethod	-	the most recently resolved concrete method implementation (CompiledMethod)
	4:	hotness			-	the hotness counter for this send (SmallInteger)
	5:	trips			-	the trip counter for this cache (SmallInteger)

The hotness counter is incremented each time the send is executed. The trip counter is incremented each time the cached class and method are replaced with new values because the cache has produced a miss.!

!TFMonomorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:00'!
defaultSize
	"Return the size of a TFSingleEntryCache object. It contains 5 entries as specified in the class comment."
	
	^ 5! !

!TFMonomorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:03'!
forSelector: aSelector resolvedClass: aClass resolvedMethod: aCompiledMethod
	"Create an instance of the cache for a given selector and most recently resolved class and method. Set hotness and trip counters to zero."
	
	^ (self new: self defaultSize)
		at: 1 put: aSelector;
		at: 2 put: aClass;
		at: 3 put: aCompiledMethod;
		at: 4 put: 0;		"hotness counter"
		at: 5 put: 0		"trip counter";
		yourself! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:05'!
hotness
	^ self at: 4! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:04'!
selector
	^ self at: 1! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:05'!
targetClass
	^ self at: 2! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:05'!
targetMethod
	^ self at: 3! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:05'!
trips
	^ self at: 5! !

Array variableSubclass: #TFPolymorphicCache
	instanceVariableNames: 'nextFreeIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TypeFeedback'!

!TFPolymorphicCache commentStamp: 'mha 6/18/2007 21:16' prior: 0!
An instance of this class represents a polymorphic cache, i.e., a send site-local cache that stores a given number of concrete target methods for the send in question. The class is a subclass of Array. Its instances are 1+(3*n) slots long, where n is the number of concrete receivers for the send.

At index 1, a polymorphic cache always contains the selector (a Symbol) denoting the message sent at this send site.

All subsequent indices are grouped in triples. Each triple contains, for the cache entry k, at the array index given below, the following data:

	k*3-1	:	the resolved target method (CompiledMethod)
	k*3		:	the concrete resolved target class (Class)
	k*3+1	:	the hotness counter for this particular cache entry (SmallInteger)

The order of concrete resolved method and class is reversed as compared to TFMonomorphicCache to allow for faster class lookup (one less arithmetic operation).

The hotness counter is maintained to allow for later optimisation (e.g., inlining). It is incremented each time the particular cache entry is hit.!

!TFPolymorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:00'!
defaultSize
	"Answer the number of entries that a newly created polymorphic cache should have."
	
	^ 4! !

!TFPolymorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:08'!
fromMonomorphicCache: monomorphicCache
	"Create a new polymorphic cache and initialise it with the data found in the monomorphic cache (a TFMonomorphicCache) passed as the argument. The argument's contents are placed in the first entry of the polymorphic cache. All other entries are left blank; they are to be filled subsequently by the virtual machine."
	
	^ (self new: self defaultSize * 3) "one cache entry consumes 3 slots"
		at: 1 put: monomorphicCache selector;
		at: 2 put: monomorphicCache targetClass;
		at: 3 put: monomorphicCache targetMethod;
		at: 4 put: monomorphicCache hotness;
		nextFreeIndex: 2;
		yourself! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:14'!
hotnessAt: k
	^ self at: (k * 3 + 1)! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:09'!
nextFreeIndex
	"Answer the next free cache entry index."
	
	^ nextFreeIndex! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/18/2007 21:10'!
nextFreeIndex: idx
	"Set the next free cache entry index."
	
	nextFreeIndex := idx! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:10'!
selector
	^ self at: 1! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:16'!
targetClassAt: k
	^ self at: k * 3! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/18/2007 21:16'!
targetMethodAt: k
	^ self at: (k * 3 - 1)! !
